﻿// A Closed-Bucket Hashtable
// 2010/06/12 - Bradley Smith

using System;
using System.Collections.Generic;

/// <summary>
/// Provides the base class for a generic implementation of a closed-bucket hashtable, in which each key may have multiple values.
/// </summary>
/// <typeparam name="TKey">The type of the keys in the collection.</typeparam>
/// <typeparam name="TValue">The type of the values in the collection.</typeparam>
public abstract class CBHTBase<TKey, TValue> : IDictionary<TKey, ICollection<TValue>>, ICollection<TValue>, IDictionary<TKey, TValue> {

	/// <summary>
	/// Returns the total number of values in the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	public int Count {
		get {
			int total = 0;
			foreach (TKey key in Keys) total += this[key].Count;
			return total;
		}
	}

	/// <summary>
	/// Gets a value indicating whether the IDictionary is read-only. This property always returns false.
	/// </summary>
	public bool IsReadOnly {
		get {
			return false;
		}
	}

	/// <summary>
	/// Returns the number of keys in the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	public abstract int KeyCount {
		get;
	}

	/// <summary>
	/// Gets a collection containing the keys in the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	public abstract ICollection<TKey> Keys {
		get;
	}

	/// <summary>
	/// Gets or sets the value(s) associated with the specified key.
	/// </summary>
	/// <param name="key">The key whose associated values are desired.</param>
	/// <returns></returns>
	public abstract ICollection<TValue> this[TKey key] {
		get;
		set;
	}

	/// <summary>
	/// Gets a collection containing the values in the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	public ICollection<TValue> Values {
		get {
			// the collection implements the necessary interface already
			return this;
		}
	}

	/// <summary>
	/// Adds the specified key/value pair to the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	/// <param name="key">The key to associate the value with.</param>
	/// <param name="value">The value to add to the collection.</param>
	public abstract void Add(TKey key, TValue value);

	/// <summary>
	/// Removes all keys and values from the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	public abstract void Clear();

	/// <summary>
	/// Determines whether the CBHTBase&lt;TKey, TValue&gt; contains a specific value.
	/// </summary>
	/// <param name="value">The value to test for.</param>
	/// <returns>True if found, otherwise false.</returns>
	public bool Contains(TValue value) {
		// sequential search
		foreach (ICollection<TValue> bucket in this) {
			if (bucket.Contains(value)) return true;
		}

		return false;
	}

	/// <summary>
	/// Determines whether the CBHTBase&lt;TKey, TValue&gt; contains the specified key/value pair.
	/// </summary>
	/// <param name="item">The key/value pair to test for.</param>
	/// <returns>True if the key/value pair is found.</returns>
	public bool Contains(KeyValuePair<TKey, TValue> item) {
		return ContainsKey(item.Key) && this[item.Key].Contains(item.Value);
	}

	/// <summary>
	/// Determines whether the CBHTBase&lt;TKey, TValue&gt; contains the specified key.
	/// </summary>
	/// <param name="key">The key to test for.</param>
	/// <returns>True if found, otherwise false.</returns>
	public abstract bool ContainsKey(TKey key);

	/// <summary>
	/// Copies all the values in the CBHTBase&lt;TKey, TValue&gt; into a compatible array, starting at the specified index.
	/// </summary>
	/// <param name="array">The array to copy the elements into.</param>
	/// <param name="arrayIndex">The index to start copying at.</param>
	public void CopyTo(TValue[] array, int arrayIndex) {
		if (array == null) throw new ArgumentNullException("array");
		if ((arrayIndex < 0) || (arrayIndex >= array.Length)) throw new ArgumentOutOfRangeException("arrayIndex");
		if (Count < (array.Length - arrayIndex)) throw new ArgumentException("Array is too short");

		foreach (TValue item in this) {
			array[arrayIndex++] = item;
		}
	}

	/// <summary>
	/// Returns an enumerator that iterates through all the values in the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	/// <returns>An enumerator for the values in the collection.</returns>
	public abstract IEnumerator<TValue> GetEnumerator();

	/// <summary>
	/// Removes the specified key from the CBHTBase&lt;TKey, TValue&gt; and all values associated with it.
	/// </summary>
	/// <param name="key">The key whose associated values will be removed.</param>
	/// <returns>True if the key is found.</returns>
	public abstract bool Remove(TKey key);

	/// <summary>
	/// Removes the specified key/value pair from the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	/// <param name="item">The key/value pair to remove.</param>
	/// <returns>True if the key/value pair is found.</returns>
	public bool Remove(KeyValuePair<TKey, TValue> item) {
		if (Contains(item)) {
			if (this[item.Key].Remove(item.Value)) {
				// if there are no more values associated with the key, remove it
				if (this[item.Key].Count == 0) Remove(item.Key);
				return true;
			}
		}

		return false;
	}

	/// <summary>
	/// Removes the first occurrence of a value in the CBHTBase&lt;TKey, TValue&gt;.
	/// </summary>
	/// <param name="value">The value to remove.</param>
	/// <returns>True if the value is found.</returns>
	public bool RemoveByValue(TValue value) {
		foreach (KeyValuePair<TKey, TValue> pair in (ICollection<KeyValuePair<TKey, TValue>>)this) {
			if (this[pair.Key].Remove(pair.Value)) {
				// if there are no more values associated with the key, remove it
				if (this[pair.Key].Count == 0) this.Remove(pair.Key);
				return true;
			}
		}

		return false;
	}

	#region IDictionary<TKey, ICollection<TValue>> Members

	void IDictionary<TKey, ICollection<TValue>>.Add(TKey key, ICollection<TValue> value) {
		// if value(s) already exist for the key, add more, otherwise create a new linked list in the bucket
		if (ContainsKey(key))
			foreach (TValue item in value) this[key].Add(item);
		else
			this[key] = value;
	}

	void ICollection<KeyValuePair<TKey, ICollection<TValue>>>.Add(KeyValuePair<TKey, ICollection<TValue>> item) {
		((IDictionary<TKey, ICollection<TValue>>)this).Add(item.Key, item.Value);
	}

	bool ICollection<KeyValuePair<TKey, ICollection<TValue>>>.Contains(KeyValuePair<TKey, ICollection<TValue>> item) {
		return ContainsKey(item.Key) && this[item.Key].Equals(item.Value);
	}

	void ICollection<KeyValuePair<TKey, ICollection<TValue>>>.CopyTo(KeyValuePair<TKey, ICollection<TValue>>[] array, int arrayIndex) {
		if (array == null) throw new ArgumentNullException("array");
		if ((arrayIndex < 0) || (arrayIndex >= array.Length)) throw new ArgumentOutOfRangeException("arrayIndex");
		if (KeyCount < (array.Length - arrayIndex)) throw new ArgumentException("Array is too short");

		foreach (KeyValuePair<TKey, ICollection<TValue>> item in (IEnumerable<KeyValuePair<TKey, ICollection<TValue>>>)this) {
			array[arrayIndex++] = item;
		}
	}

	int ICollection<KeyValuePair<TKey, ICollection<TValue>>>.Count {
		get {
			return KeyCount;
		}
	}

	IEnumerator<KeyValuePair<TKey, ICollection<TValue>>> IEnumerable<KeyValuePair<TKey, ICollection<TValue>>>.GetEnumerator() {
		foreach (TKey key in Keys) {
			yield return new KeyValuePair<TKey, ICollection<TValue>>(key, this[key]);
		}
	}

	bool ICollection<KeyValuePair<TKey, ICollection<TValue>>>.Remove(KeyValuePair<TKey, ICollection<TValue>> item) {
		if (((ICollection<KeyValuePair<TKey, ICollection<TValue>>>)this).Contains(item))
			return Remove(item.Key);
		else
			return false;
	}

	bool IDictionary<TKey, ICollection<TValue>>.TryGetValue(TKey key, out ICollection<TValue> value) {
		// lack of covariance/contravariance requires declaration of a more specific variable
		if (ContainsKey(key)) {
			value = this[key];
			return true;
		}

		value = null;
		return false;
	}

	ICollection<ICollection<TValue>> IDictionary<TKey, ICollection<TValue>>.Values {
		get {
			// lack of covariance/contravariance means that we have to return a separate collection
			List<ICollection<TValue>> values = new List<ICollection<TValue>>(KeyCount);
			foreach (TKey key in Keys) values.Add(this[key]);
			return values.AsReadOnly();
		}
	}

	#endregion

	#region ICollection<TValue> Members

	void ICollection<TValue>.Add(TValue item) {
		// read only
		throw new NotSupportedException();
	}

	void ICollection<TValue>.Clear() {
		// read only
		throw new NotSupportedException();
	}

	bool ICollection<TValue>.IsReadOnly {
		get {
			// can't add manipulate the values without the keys, so make read only
			return true;
		}
	}

	bool ICollection<TValue>.Remove(TValue item) {
		// read only
		throw new NotSupportedException();
	}

	#endregion

	#region IDictionary<TKey, TValue> Members

	TValue IDictionary<TKey, TValue>.this[TKey key] {
		get {
			TValue attempt;
			if (((IDictionary<TKey, TValue>)this).TryGetValue(key, out attempt))
				return attempt;
			else
				throw new InvalidOperationException();		// cannot return a single value
		}
		set {
			// set a single value, clearing the collection if it exists
			if (ContainsKey(key))
				this[key].Clear();
			else
				this[key] = new TValue[] { };

			this[key].Add(value);
		}
	}

	void ICollection<KeyValuePair<TKey, TValue>>.Add(KeyValuePair<TKey, TValue> item) {
		Add(item.Key, item.Value);
	}

	void ICollection<KeyValuePair<TKey, TValue>>.CopyTo(KeyValuePair<TKey, TValue>[] array, int arrayIndex) {
		if (array == null) throw new ArgumentNullException("array");
		if ((arrayIndex < 0) || (arrayIndex >= array.Length)) throw new ArgumentOutOfRangeException("arrayIndex");
		if (Count < (array.Length - arrayIndex)) throw new ArgumentException("Array is too short");

		foreach (KeyValuePair<TKey, TValue> item in (IEnumerable<KeyValuePair<TKey, TValue>>)this) {
			array[arrayIndex++] = item;
		}
	}

	IEnumerator<KeyValuePair<TKey, TValue>> IEnumerable<KeyValuePair<TKey, TValue>>.GetEnumerator() {
		foreach (KeyValuePair<TKey, ICollection<TValue>> pair in (IDictionary<TKey, ICollection<TValue>>)this) {
			foreach (TValue value in pair.Value) yield return new KeyValuePair<TKey, TValue>(pair.Key, value);
		}
	}

	bool IDictionary<TKey, TValue>.TryGetValue(TKey key, out TValue value) {
		// get the one and only value, or fail
		if (ContainsKey(key) && (this[key].Count == 1)) {
			foreach (TValue tmp in this[key]) {
				value = tmp;
				return true;
			}
		}

		value = default(TValue);
		return false;
	}

	#endregion

	#region System.Collections.IEnumerable Members

	System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator() {
		return GetEnumerator();
	}

	#endregion
}