﻿// A Sorted Closed-Bucket Hashtable
// 2010/06/12 - Bradley Smith

using System;
using System.Collections.Generic;

/// <summary>
/// Provides a generic implementation of a closed-bucket hashtable, in which each key may have multiple values. 
/// This implementation uses System.Collections.Generic.SortedList&lt;TKey, TValue&gt; as a container, with each 
/// bucket being represented by a binary search tree.
/// </summary>
/// <typeparam name="TKey">The type of the keys in the collection.</typeparam>
/// <typeparam name="TValue">The type of the values in the collection.</typeparam>
class SortedCBHT<TKey, TValue> : CBHTBase<TKey, TValue> {

	SortedList<TKey, BST<TValue>> mInnerDic;

	/// <summary>
	/// Initialises a new instance of the SortedSortedCBHT&lt;TKey, TValue&gt; class.
	/// </summary>
	public SortedCBHT() {
		mInnerDic = new SortedList<TKey, BST<TValue>>();
	}

	/// <summary>
	/// Initialises a new instance of the SortedSortedCBHT&lt;TKey, TValue&gt; class, using the specified initial capacity.
	/// </summary>
	/// <param name="capacity">The initial capacity of the inner System.Collections.Generic.SortedList&lt;TKey, TValue&gt;.</param>
	public SortedCBHT(int capacity) {
		mInnerDic = new SortedList<TKey, BST<TValue>>(capacity);
	}

	/// <summary>
	/// Returns the number of keys in the SortedCBHT&lt;TKey, TValue&gt;.
	/// </summary>
	public override int KeyCount {
		get {
			return mInnerDic.Count;
		}
	}

	/// <summary>
	/// Gets a collection containing the keys in the SortedCBHT&lt;TKey, TValue&gt;.
	/// </summary>
	public override ICollection<TKey> Keys {
		get {
			return mInnerDic.Keys;
		}
	}

	/// <summary>
	/// Gets or sets the value(s) associated with the specified key.
	/// </summary>
	/// <param name="key">The key whose associated values are desired.</param>
	/// <returns></returns>
	public override ICollection<TValue> this[TKey key] {
		get {
			return mInnerDic[key];
		}
		set {
			if (mInnerDic.ContainsKey(key)) {
				mInnerDic[key].Clear();
				foreach (TValue item in value) mInnerDic[key].Add(item);
			}
			else {
				mInnerDic[key] = new BST<TValue>(value);
			}
		}
	}

	/// <summary>
	/// Adds the specified key/value pair to the SortedCBHT&lt;TKey, TValue&gt;.
	/// </summary>
	/// <param name="key">The key to associate the value with.</param>
	/// <param name="value">The value to add to the collection.</param>
	public override void Add(TKey key, TValue value) {
		if (!mInnerDic.ContainsKey(key)) mInnerDic[key] = new BST<TValue>();
		mInnerDic[key].Add(value);
	}

	/// <summary>
	/// Removes all keys and values from the SortedCBHT&lt;TKey, TValue&gt;.
	/// </summary>
	public override void Clear() {
		mInnerDic.Clear();
	}

	/// <summary>
	/// Determines whether the SortedCBHT&lt;TKey, TValue&gt; contains the specified key.
	/// </summary>
	/// <param name="key">The key to test for.</param>
	/// <returns>True if found, otherwise false.</returns>
	public override bool ContainsKey(TKey key) {
		return mInnerDic.ContainsKey(key);
	}

	/// <summary>
	/// Returns an enumerator that iterates through all the values in the SortedCBHT&lt;TKey, TValue&gt;.
	/// </summary>
	/// <returns>An enumerator for the values in the collection.</returns>
	public override IEnumerator<TValue> GetEnumerator() {
		foreach (ICollection<TValue> values in mInnerDic.Values) {
			foreach (TValue value in values) yield return value;
		}
	}

	/// <summary>
	/// Removes the specified key from the SortedCBHT&lt;TKey, TValue&gt; and all values associated with it.
	/// </summary>
	/// <param name="key">The key whose associated values will be removed.</param>
	/// <returns>True if the key is found.</returns>
	public override bool Remove(TKey key) {
		return mInnerDic.Remove(key);
	}
}
